require 'tk'
require 'rexml/document' 
include REXML

class SequenceFlow
	attr_accessor :sourceid, :info, :points, :elementid, :highlight

	def initialize(sourceid, info, canvas, points)
		@sourceid = sourceid 
		@info = info
		@canvas = canvas
		@points = points
		create
	end
	
	def create	
		line = TkcLine.new(@canvas,@points)
		line.smooth = true
		line.arrow = "last"
		oldcolor = "black"				
		line.bind("Enter") { oldcolor = line.fill; line.fill = "gray"; @info.text='id=e'+@sourceid; }
		line.bind("Leave") { line.fill = oldcolor; @info.text='-'}
		@elementid = line
		@highlight = line
	end
	
end

class Task

	attr_accessor :sourceid, :info, :canvas, :name, :x, :y, :w, :h, :elementid, :highlight

	def initialize(sourceid, info, canvas,name,x,y,w,h)
		@sourceid = sourceid	
		@info = info
		@canvas = canvas
		@name = name
		@x = x
		@y = y
		@w = w
		@h = h
		create
	end
	
	def create
		tag = TkcTag.new(@canvas)
		rect = TkcRectangle.new(@canvas,@x,@y,@x+@w,@y+@h, 'tag'=>tag)
		rect.fill = "white"
		TkcText.new(@canvas,@x+@w/2,@y+@h/2,'text'=>@name,'anchor'=>'c', 'tag'=>tag)
		oldcolor = "white"		
		tag.bind("Enter") { oldcolor = rect.fill; rect.fill = "gray"; @info.text='id=N'+@sourceid;  }
		tag.bind("Leave") { rect.fill = oldcolor; @info.text='-' }
		@elementid = tag
		@highlight = rect
	end
	
end

class Event

	attr_accessor :sourceid, :info, :canvas, :x, :y, :w, :h, :style, :elementid, :highlight
	
	def initialize(sourceid,info,canvas,x,y,w,h,style)
		@sourceid = sourceid	
		@info = info
		@canvas = canvas
		@x = x
		@y = y
		@w = w
		@h = h
		@style = style
		create
	end
	
	def create		
		tag = TkcTag.new(@canvas)	
		oval = TkcOval.new(@canvas,@x,@y,@x+@w,@y+@h, 'tag'=>tag)
		oval.fill = "white"
		
		case style
			when "Start Event":
			
			when "Intermediate Event":
				oval2 = TkcOval.new(@canvas,@x+3,@y+3,@x+@w-3,@y+@h-3, 'tag'=>tag)		
					
			when "End Event":
				oval.width = "3"				
			else
				TkcText.new(@canvas,@x+@w/2,@y+@h/2,'text'=>'?','anchor'=>'c', 'tag'=>tag)						
		end
		
		oldcolor = "white"		
		tag.bind("Enter") { oldcolor = oval.fill; oval.fill = "gray"; @info.text='id=N'+@sourceid; }
		tag.bind("Leave") { oval.fill = oldcolor; @info.text='-' }
		@elementid = tag
		@highlight = oval
		
	end	

end

class Gateway
	
	attr_accessor :sourceid, :info, :canvas, :x, :y, :w, :h, :style, :elementid, :highlight
	
	def initialize(sourceid,info,canvas,x,y,w,h,style)
		@sourceid = sourceid	
		@info = info
		@canvas = canvas
		@x = x
		@y = y
		@w = w
		@h = h
		@style = style
		create
	end
	
	def create		
		tag = TkcTag.new(@canvas)	
		poly = TkcPolygon.new(@canvas, @x+(@w/2), @y, @x,@y+(@h/2), @x+(@w/2), @y+@h, @x+@w, @y+(@h/2),
							 'tag'=>tag)
		poly.fill = "white"
		poly.outline = "black"
		
		case style
			when "XOR Gateway":
				TkcText.new(@canvas,@x+@w/2,@y+@h/2,'text'=>'X','anchor'=>'c', 'tag'=>tag)	
			when "AND Gateway":
				TkcText.new(@canvas,@x+@w/2,@y+@h/2,'text'=>'+','anchor'=>'c', 'tag'=>tag)	
			when "OR Gateway":
				TkcText.new(@canvas,@x+@w/2,@y+@h/2,'text'=>'O','anchor'=>'c', 'tag'=>tag)		
			else
				TkcText.new(@canvas,@x+@w/2,@y+@h/2,'text'=>'?','anchor'=>'c', 'tag'=>tag)						
		end
		
		oldcolor = "white"		
		tag.bind("Enter") { oldcolor = poly.fill; poly.fill = "gray"; @info.text='id=N'+@sourceid; }
		tag.bind("Leave") { poly.fill = oldcolor; @info.text='-' }
		@highlight = poly
		@elementid = tag
	end	
	

end

class BPMNView

	attr_accessor :canvas, :elements, :info

	def initialize(xml,id)
	
		print "ID " + id.to_s + " "
	
		Tk.restart
		@root = TkRoot.new { title 'BPMN View for process id = '+id.to_s }
		#@bar = TkScrollbar.new(@root, 'orient'=>'hor').pack('side'=>'bottom', 'fill'=>'x')
		@canvas = TkCanvas.new(@root, 'width'=>640, 'height'=>480).pack('fill'=>'both', 'expand'=>true)
		@info = TkLabel.new(@root).pack('fill'=>'x', 'expand'=>true)
		@info.text = "-"
		#@info.background = "gray"
		#@bar.command(proc { |args| canvas.xview(*args) })
		#@canvas.xscrollcommand(proc { |first, last|@bar.set(first, last) })
		@elements = Array.new
		@minX = @minY = 10000.0
		@maxX = @maxY = @offsetX = @offsetY =  @factor = 0.0
		
		TkcText.new(@canvas,5,5,'text'=>'Active','anchor'=>'nw', 'fill'=>'green')		
		TkcText.new(@canvas,5,20,'text'=>'Internal','anchor'=>'nw', 'fill'=>'orange')		
		TkcText.new(@canvas,5,35,'text'=>'Blocked','anchor'=>'nw', 'fill'=>'red')		
		
		load(xml,id)
	
		return @canvas
	end
	
	#
	# Loads a BPMN diagram with the given id from an XML file
	#
	def load(xml,id)
 	  	
 	  	#
 	  	# Process task
 	  	#
	  	xml.elements.each("model/process[@id=#{id}]/node[@type='Task']") do |n|
	  		taskName = n.attributes['name'].to_s
	  		layout = XPath.first(xml,"model/layout/node[@id=#{n.attributes['id']}]")
			sourceid,x,y,w,h = _getElementPositions(layout)
	  		@elements << Task.new(sourceid,@info,@canvas, taskName, x, y, w, h)
	  		end
	  			  		
	  	#
	  	# Process event
	  	#
	  	xml.elements.each("model/process[@id=#{id}]/node[contains(@type,'Event')]") do |n|
			layout = XPath.first(xml,"model/layout/node[@id=#{n.attributes['id']}]")			
	  		sourceid,x,y,w,h = _getElementPositions(layout)
			@elements << Event.new(sourceid,@info,@canvas, x, y, w, h, n.attributes['type'])
	  		end	  	
	  		
	  	#
	  	# Process gateways
	  	#	
	  	xml.elements.each("model/process[@id=#{id}]/node[contains(@type,'Gateway')]") do |n|
	  		layout = XPath.first(xml,"model/layout/node[@id=#{n.attributes['id']}]")			
	  		sourceid,x,y,w,h = _getElementPositions(layout)
			@elements << Gateway.new(sourceid,@info,@canvas, x, y, w, h, n.attributes['type'])
	  		end	  		
	  		
	  	#
	  	# Process sequence flows
	  	#  	
	  	xml.elements.each("model/process[@id=#{id}]/flow[@type='Sequence Flow']") do |n|
  	
	  		layout = XPath.first(xml,"model/layout/flow[@id=#{n.attributes['id']}]")			
	  		sourceid = layout.attributes['id'].to_s
	  		# Get all points
	  		points = []
	  		pos = 1
	  		while true 
	  			x = layout.attributes['x' + pos.to_s]
				y = layout.attributes['y' + pos.to_s]
				break if !x 
				points << x.to_f
				points << y.to_f
				pos += 1
		  		end
	  		
			@elements << SequenceFlow.new(sourceid,@info,@canvas, points)
	  		end	 	  	
	  		
	  	updatePositions  	
	end
	
	def _getElementPositions(layout)
		if !layout					
	  		raise "NO LAYOUT INFORMATION FOUND!"  	
			end
	
		sourceid = layout.attributes['id'].to_s
	  	x = layout.attributes['x'].to_f
  		y = layout.attributes['y'].to_f
  		w = layout.attributes['w'].to_f
  		h = layout.attributes['h'].to_f
  		@minX = x if x < @minX
  		@maxX = x + w if (x + w) > @maxX
  		@minY = y if y < @minY
  		@maxY = y + h if (y + h) > @maxY
  		return sourceid,x,y,w,h
	end
	
	def clearHighlights()
		@elements.each do |e|  	  
			case e
				when SequenceFlow:
					e.highlight.fill = "black"
				else				
		  			e.highlight.fill = "white"
		  		end
  		end		
	end
	
	def highlight(id,color)
		@elements.each do |e|  	
	  		if e.sourceid.to_s == id.to_s	  		
	  			e.highlight.fill = color
	  			end
	  		end		
	end	
	
	def updatePositions
		#puts "BOUNDING BOX = (" + @minX.to_s + "," + @minY.to_s + ")->(" + @maxX.to_s + "," + @maxY.to_s + ")"
 	
	  	# Calculate factor if !!!
	  	if (@maxX - @minX) > (@canvas.width-20)
	  		@factor = (@canvas.width-20)/(@maxX - @minX)
	  		else
	  		@factor = 1.0
	 		end	 
	  		
	  	# Calculate offset
	  	@offsetX = (@canvas.width/2) - ((@maxX - @minX)/2) - @minX 
	  	@offsetY = (@canvas.height/2) - ((@maxY - @minY)/2) - @minY 
			  	
	  	#puts "offsetX = " + @offsetX.to_s + " offsetY = " + @offsetY.to_s + " factor = " + @factor.to_s
	  	
	  	# Update the positions of all elements
	  	@elements.each do |e|
	  		@canvas.move(e.elementid,@offsetX,@offsetY)
	  		@canvas.scale(e.elementid,@canvas.width/2,@canvas.height/2,@factor,@factor)
	  		end

	end
	
	def kill
		@root.destroy
	end	
	
	def run
		Tk.mainloop
	end	

end

#view = BPMNView.new(Document.new(File.open('test.xml')),1)
#c = Event.new(view,10,10,20,20)
#c.draw
#view.run
